#!/usr/bin/env python3

import os
import flatbuffers
import numpy as np
import keras
from keras import layers, ops
import tensorflow as tf
from tensorflow.lite.python import schema_py_generated as tflite_schema


BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def convert_keras(model, out_name):
    blob = tf.lite.TFLiteConverter.from_keras_model(model).convert()
    out_path = os.path.join(BASE_DIR, out_name)
    with open(out_path, "wb") as f:
        f.write(blob)
    return out_path


def convert_concrete(concrete_fn, out_name):
    converter = tf.lite.TFLiteConverter.from_concrete_functions(
        [concrete_fn], trackable_obj=None
    )
    blob = converter.convert()
    out_path = os.path.join(BASE_DIR, out_name)
    with open(out_path, "wb") as f:
        f.write(blob)
    return out_path


def dims_to_str(shape_sig):
    return ",".join("-1" if int(v) < 0 else str(int(v)) for v in shape_sig)


def dtype_to_modelinfo(dt):
    if dt == np.float32:
        return "float32"
    if dt == np.uint8:
        return "uint8"
    if dt == np.int8:
        return "int8"
    if dt == np.int16:
        return "int16"
    if dt == np.int32:
        return "int32"
    if dt == np.int64:
        return "int64"
    raise ValueError(f"Unsupported dtype {dt}")


def write_modelinfo(model_path, include_ranges=True, per_output_dims_order=None):
    if per_output_dims_order is None:
        per_output_dims_order = {}

    itp = tf.lite.Interpreter(model_path=model_path)
    ins = itp.get_input_details()
    outs = itp.get_output_details()

    group_id = os.path.basename(model_path).replace(".tflite", "") + "-group"
    lines = ["[modelinfo]", "version=1.0", f"group-id={group_id}", ""]

    for i, d in enumerate(ins):
        name = d["name"]
        dtype = dtype_to_modelinfo(d["dtype"])
        dims = dims_to_str(d.get("shape_signature", d["shape"]))
        lines += [
            f"[{name}]",
            f"id=input-{i}",
            f"type={dtype}",
            f"dims={dims}",
            "dir=input",
        ]
        if include_ranges:
            lines.append("ranges=0.0,255.0")
        lines.append("")

    for i, d in enumerate(outs):
        name = d["name"]
        dtype = dtype_to_modelinfo(d["dtype"])
        dims = dims_to_str(d.get("shape_signature", d["shape"]))
        lines += [
            f"[{name}]",
            f"id=output-{i}",
            f"type={dtype}",
            f"dims={dims}",
            "dir=output",
        ]
        if i in per_output_dims_order:
            lines.append(f"dims-order={per_output_dims_order[i]}")
        lines.append("")

    with open(model_path + ".modelinfo", "w", encoding="utf-8") as f:
        f.write("\n".join(lines).rstrip() + "\n")


def make_offset_input_model(out_name, input_name, input_shape):
    output_size = int(np.prod(input_shape))
    output_shape = [1, output_size]

    model = tflite_schema.ModelT()
    model.version = 3
    model.description = (
        "Input tensor intentionally not at index 0 for inference testing"
    )

    buf0 = tflite_schema.BufferT()
    buf0.data = None
    shape_const = np.array(output_shape, dtype=np.int32)
    buf1 = tflite_schema.BufferT()
    buf1.data = list(shape_const.tobytes())
    buf2 = tflite_schema.BufferT()
    buf2.data = None
    buf3 = tflite_schema.BufferT()
    buf3.data = None
    buf4 = tflite_schema.BufferT()
    buf4.data = None
    model.buffers = [buf0, buf1, buf2, buf3, buf4]

    cast_opcode = tflite_schema.OperatorCodeT()
    cast_opcode.builtinCode = tflite_schema.BuiltinOperator.CAST
    cast_opcode.deprecatedBuiltinCode = tflite_schema.BuiltinOperator.CAST
    reshape_opcode = tflite_schema.OperatorCodeT()
    reshape_opcode.builtinCode = tflite_schema.BuiltinOperator.RESHAPE
    reshape_opcode.deprecatedBuiltinCode = tflite_schema.BuiltinOperator.RESHAPE
    model.operatorCodes = [cast_opcode, reshape_opcode]

    subgraph = tflite_schema.SubGraphT()
    subgraph.name = "main"

    t_shape = tflite_schema.TensorT()
    t_shape.name = "reshape_shape"
    t_shape.shape = np.array([2], dtype=np.int32)
    t_shape.type = tflite_schema.TensorType.INT32
    t_shape.buffer = 1

    t_cast = tflite_schema.TensorT()
    t_cast.name = "cast_output"
    t_cast.shape = np.array(input_shape, dtype=np.int32)
    t_cast.type = tflite_schema.TensorType.FLOAT32
    t_cast.buffer = 3

    t_input = tflite_schema.TensorT()
    t_input.name = input_name
    t_input.shape = np.array(input_shape, dtype=np.int32)
    t_input.type = tflite_schema.TensorType.UINT8
    t_input.buffer = 2

    t_output = tflite_schema.TensorT()
    t_output.name = "output_flat_f32"
    t_output.shape = np.array(output_shape, dtype=np.int32)
    t_output.type = tflite_schema.TensorType.FLOAT32
    t_output.buffer = 4

    # Keep input tensor at global tensor index 2 on purpose.
    subgraph.tensors = [t_shape, t_cast, t_input, t_output]
    subgraph.inputs = np.array([2], dtype=np.int32)
    subgraph.outputs = np.array([3], dtype=np.int32)

    cast_op = tflite_schema.OperatorT()
    cast_op.opcodeIndex = 0
    cast_op.inputs = np.array([2], dtype=np.int32)
    cast_op.outputs = np.array([1], dtype=np.int32)
    cast_opts = tflite_schema.CastOptionsT()
    cast_opts.inDataType = tflite_schema.TensorType.UINT8
    cast_opts.outDataType = tflite_schema.TensorType.FLOAT32
    cast_op.builtinOptionsType = tflite_schema.BuiltinOptions.CastOptions
    cast_op.builtinOptions = cast_opts

    reshape_op = tflite_schema.OperatorT()
    reshape_op.opcodeIndex = 1
    reshape_op.inputs = np.array([1, 0], dtype=np.int32)
    reshape_op.outputs = np.array([3], dtype=np.int32)
    reshape_opts = tflite_schema.ReshapeOptionsT()
    reshape_opts.newShape = np.array(output_shape, dtype=np.int32)
    reshape_op.builtinOptionsType = tflite_schema.BuiltinOptions.ReshapeOptions
    reshape_op.builtinOptions = reshape_opts

    subgraph.operators = [cast_op, reshape_op]
    model.subgraphs = [subgraph]

    builder = flatbuffers.Builder(1024)
    model_offset = model.Pack(builder)
    builder.Finish(model_offset, b"TFL3")
    out_path = os.path.join(BASE_DIR, out_name)
    with open(out_path, "wb") as f:
        f.write(bytes(builder.Output()))

    return out_path


def main():
    for n in os.listdir(BASE_DIR):
        if n.endswith(".tflite") or n.endswith(".modelinfo"):
            os.unlink(os.path.join(BASE_DIR, n))

    inp = keras.Input(shape=(4, 4, 3), batch_size=1, dtype="uint8", name="input_u8")
    x = layers.Lambda(lambda t: ops.cast(t, "float32"))(inp)
    out = layers.Reshape((48,), name="output_flat_f32")(x)
    p = convert_keras(keras.Model(inp, out), "flatten_uint8in_float32out.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(4, 4, 3), batch_size=1, dtype="float32", name="input_f32")
    out = layers.Reshape((48,), name="output_flat_f32")(inp)
    p = convert_keras(keras.Model(inp, out), "flatten_float32in_float32out.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(4, 4, 3), batch_size=1, dtype="float32", name="input_f32")
    x = layers.Lambda(lambda t: ops.clip(t, 0.0, 255.0))(inp)
    out = layers.Lambda(lambda t: ops.cast(t, "uint8"), name="output_u8")(x)
    p = convert_keras(keras.Model(inp, out), "uint8out.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(4, 4, 3), batch_size=1, dtype="float32", name="input_f32")
    x = layers.Lambda(lambda t: ops.clip(t, -128.0, 127.0))(inp)
    out = layers.Lambda(lambda t: ops.cast(t, "int8"), name="output_i8")(x)
    p = convert_keras(keras.Model(inp, out), "int8out.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(4, 4, 3), dtype="float32", name="input_dyn")
    out = layers.Reshape((48,), name="output_dyn")(inp)
    p = convert_keras(keras.Model(inp, out), "dynamic_batch.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(4, 4, 3), batch_size=1, dtype="float32", name="input_f32")
    out1 = layers.Reshape((48,), name="out1")(inp)  # (1, 48) row-major - will be output-0
    x = layers.Permute((3, 1, 2))(inp)  # Transpose to (1, 3, 4, 4) - channels first
    out0 = layers.Reshape((3, 16), name="out0")(x)  # (3, 16) col-major - will be output-1
    p = convert_keras(keras.Model(inp, [out1, out0]), "multi_output.tflite")
    write_modelinfo(p, per_output_dims_order={1: "col-major"})

    inp = keras.Input(shape=(4, 4), batch_size=3, dtype="float32", name="input_3d")
    out = layers.Reshape((16,), name="output_3d")(inp)
    p = convert_keras(keras.Model(inp, out), "flatten_3d_float32.tflite")
    write_modelinfo(p)

    inp0 = keras.Input(shape=(4, 4, 3), batch_size=1, dtype="float32", name="input0")
    inp1 = keras.Input(shape=(4, 4, 3), batch_size=1, dtype="float32", name="input1")
    out = layers.Add(name="output_add")([inp0, inp1])
    p = convert_keras(keras.Model([inp0, inp1], out), "multi_input_two_tensors.tflite")
    write_modelinfo(p)

    @tf.function(input_signature=[tf.TensorSpec(shape=[48], dtype=tf.float32, name="input_1d")])
    def model_1d(x):
        return tf.identity(x, name="output_1d")

    p = convert_concrete(model_1d.get_concrete_function(), "invalid_dims_1d.tflite")
    write_modelinfo(p)

    @tf.function(input_signature=[tf.TensorSpec(shape=[1, 2, 2, 2, 3], dtype=tf.float32, name="input_5d")])
    def model_5d(x):
        return tf.identity(x, name="output_5d")

    p = convert_concrete(model_5d.get_concrete_function(), "invalid_dims_5d.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(4, 4, 2), batch_size=1, dtype="float32", name="input_ch2")
    out = layers.Lambda(lambda t: t, name="output_ch2")(inp)
    p = convert_keras(keras.Model(inp, out), "invalid_dims_channels_2.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(3, 4, 4), batch_size=1, dtype="float32", name="input_chw")
    out = layers.Lambda(lambda t: t, name="output_chw")(inp)
    p = convert_keras(keras.Model(inp, out), "planar_chw.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(3, 4, 4), batch_size=1, dtype="uint8", name="input_chw_u8")
    x = layers.Lambda(lambda t: ops.cast(t, "float32"))(inp)
    out = layers.Reshape((48,), name="output_flat_f32")(x)
    p = convert_keras(
        keras.Model(inp, out), "planar_chw_uint8in_float32out.tflite"
    )
    write_modelinfo(p)

    p = make_offset_input_model(
        "offset_input_planar_chw_uint8in_float32out.tflite",
        "input_chw_u8",
        [1, 3, 4, 4],
    )
    write_modelinfo(p)

    p = make_offset_input_model(
        "offset_input_hwc_uint8in_float32out.tflite",
        "input_hwc_u8",
        [1, 4, 4, 3],
    )
    write_modelinfo(p)

    @tf.function(input_signature=[tf.TensorSpec(shape=[4, 4], dtype=tf.float32, name="input_gray2d")])
    def model_gray2d(x):
        return tf.identity(x, name="output_gray2d")

    p = convert_concrete(model_gray2d.get_concrete_function(), "grayscale_2d.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(4, 4, 1), batch_size=1, dtype="float32", name="input_gray4d")
    out = layers.Lambda(lambda t: t, name="output_gray4d")(inp)
    p = convert_keras(keras.Model(inp, out), "grayscale_4d.tflite")
    write_modelinfo(p)

    inp = keras.Input(shape=(4, 4, 1), batch_size=1, dtype="uint8", name="input_gray4d_u8")
    out = layers.Lambda(lambda t: ops.cast(t, "float32"), name="output_gray4d_f32")(inp)
    p = convert_keras(keras.Model(inp, out), "grayscale_uint8in_float32out.tflite")
    write_modelinfo(p)

    with open(os.path.join(BASE_DIR, "corrupt_model.tflite"), "wb") as f:
        f.write(b"not-a-valid-tflite-model")

    print("Generated models in", BASE_DIR)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os
import sys
import onnx
from onnx import helper, TensorProto
import onnxruntime as ort

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def dtype_to_modelinfo(onnx_dtype):
    """Map ONNX/ONNXRuntime dtype to modelinfo string."""
    dtype_map = {
        TensorProto.FLOAT: "float32",
        TensorProto.INT32: "int32",
        TensorProto.UINT8: "uint8",
        TensorProto.INT8: "int8",
        TensorProto.DOUBLE: "float64",
        "tensor(float)": "float32",
        "tensor(int32)": "int32",
        "tensor(uint8)": "uint8",
        "tensor(int8)": "int8",
        "tensor(double)": "float64",
    }
    return dtype_map.get(onnx_dtype, str(onnx_dtype))


def dims_to_str(shape):
    """Convert shape list to modelinfo dims string (-1 for dynamic)."""
    return ",".join("-1" if (v is None or v < 0) else str(v) for v in shape)


def write_modelinfo(onnx_path, per_output_dims_order=None):
    """
    Introspect ONNX model and generate .modelinfo file.
    Uses onnxruntime to extract tensor info.
    """
    if per_output_dims_order is None:
        per_output_dims_order = {}

    session = ort.InferenceSession(onnx_path, providers=['CPUExecutionProvider'])
    input_names = [inp.name for inp in session.get_inputs()]
    input_types = [inp.type for inp in session.get_inputs()]
    input_shapes = [inp.shape for inp in session.get_inputs()]

    output_names = [out.name for out in session.get_outputs()]
    output_types = [out.type for out in session.get_outputs()]
    output_shapes = [out.shape for out in session.get_outputs()]

    group_id = os.path.basename(onnx_path).replace(".onnx", "") + "-group"
    lines = ["[modelinfo]", "version=1.0", f"group-id={group_id}", ""]

    # Input section (required for onnxinference)
    for i, (name, onnx_type, shape) in enumerate(zip(input_names, input_types, input_shapes)):
        dtype_str = dtype_to_modelinfo(onnx_type)
        dims_str = dims_to_str(shape)
        lines += [
            f"[{name}]",
            f"id=input-{i}",
            f"type={dtype_str}",
            f"dims={dims_str}",
            "dir=input",
            "ranges=0.0,255.0;0.0,255.0;0.0,255.0",  # R,G,B ranges for 3 channels
            ""
        ]

    # Output section
    for i, (name, onnx_type, shape) in enumerate(zip(output_names, output_types, output_shapes)):
        dtype_str = dtype_to_modelinfo(onnx_type)
        dims_str = dims_to_str(shape)
        lines += [
            f"[{name}]",
            f"id=output-{i}",
            f"type={dtype_str}",
            f"dims={dims_str}",
            "dir=output",
        ]
        if i in per_output_dims_order:
            lines.append(f"dims-order={per_output_dims_order[i]}")
        lines.append("")

    modelinfo_path = onnx_path + ".modelinfo"
    with open(modelinfo_path, "w") as f:
        f.write("\n".join(lines).rstrip() + "\n")


def make_model_flatten_f32_f32():
    """[1,4,4,3] float32 → reshape → [1,48] float32"""
    input_name = "input_f32"
    output_name = "output_flat_f32"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [1, 4, 4, 3]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [1, 48]
    )

    reshape_node = helper.make_node(
        "Reshape",
        inputs=[input_name, "shape_const"],
        outputs=[output_name],
    )

    shape_const = helper.make_tensor(
        name="shape_const",
        data_type=TensorProto.INT64,
        dims=[2],
        vals=[1, 48]
    )

    graph = helper.make_graph(
        [reshape_node],
        "flatten_f32_f32",
        [input_tensor],
        [output_tensor],
        [shape_const]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    model.ir_version = 10
    return model


def make_model_uint8_to_f32():
    """[1,4,4,3] uint8 → cast float32 → reshape → [1,48] float32"""
    input_name = "input_u8"
    cast_output = "cast_out"
    output_name = "output_flat_f32"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.UINT8, [1, 4, 4, 3]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [1, 48]
    )

    cast_node = helper.make_node(
        "Cast",
        inputs=[input_name],
        outputs=[cast_output],
        to=TensorProto.FLOAT
    )

    reshape_node = helper.make_node(
        "Reshape",
        inputs=[cast_output, "shape_const"],
        outputs=[output_name],
    )

    shape_const = helper.make_tensor(
        name="shape_const",
        data_type=TensorProto.INT64,
        dims=[2],
        vals=[1, 48]
    )

    graph = helper.make_graph(
        [cast_node, reshape_node],
        "uint8_to_f32",
        [input_tensor],
        [output_tensor],
        [shape_const]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_int32_out():
    """[1,4,4,3] float32 → cast int32 → [1,4,4,3] int32"""
    input_name = "input_f32"
    output_name = "output_i32"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [1, 4, 4, 3]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.INT32, [1, 4, 4, 3]
    )

    cast_node = helper.make_node(
        "Cast",
        inputs=[input_name],
        outputs=[output_name],
        to=TensorProto.INT32
    )

    graph = helper.make_graph(
        [cast_node],
        "int32_out",
        [input_tensor],
        [output_tensor]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_dynamic_batch():
    """[-1,4,4,3] float32 → reshape → [-1,48] float32"""
    input_name = "input_dyn"
    output_name = "output_dyn"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [-1, 4, 4, 3]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [-1, 48]
    )

    # Shape computation: [-1, 48]
    reshape_node = helper.make_node(
        "Reshape",
        inputs=[input_name, "shape_const"],
        outputs=[output_name],
    )

    shape_const = helper.make_tensor(
        name="shape_const",
        data_type=TensorProto.INT64,
        dims=[2],
        vals=[-1, 48]
    )

    graph = helper.make_graph(
        [reshape_node],
        "dynamic_batch",
        [input_tensor],
        [output_tensor],
        [shape_const]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_multi_output():
    """[1,4,4,3] float32 → two outputs: [1,48] and [1,3,16]"""
    input_name = "input_f32"
    out1_name = "out1_flat"      # [1,48] row-major
    out0_name = "out0_chw"       # [1,3,16] col-major (from transpose+reshape)

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [1, 4, 4, 3]
    )
    out1_tensor = helper.make_tensor_value_info(
        out1_name, TensorProto.FLOAT, [1, 48]
    )
    out0_tensor = helper.make_tensor_value_info(
        out0_name, TensorProto.FLOAT, [1, 3, 16]
    )

    # Output 1: flatten to [1,48]
    reshape1 = helper.make_node(
        "Reshape",
        inputs=[input_name, "shape1"],
        outputs=[out1_name],
    )

    # Output 0: transpose to [1,3,4,4], then reshape to [1,3,16]
    transpose = helper.make_node(
        "Transpose",
        inputs=[input_name],
        outputs=["transpose_out"],
        perm=[0, 3, 1, 2]  # BHWC -> BCHW
    )

    reshape0 = helper.make_node(
        "Reshape",
        inputs=["transpose_out", "shape0"],
        outputs=[out0_name],
    )

    shape1_const = helper.make_tensor("shape1", TensorProto.INT64, [2], [1, 48])
    shape0_const = helper.make_tensor("shape0", TensorProto.INT64, [3], [1, 3, 16])

    graph = helper.make_graph(
        [reshape1, transpose, reshape0],
        "multi_output",
        [input_tensor],
        [out1_tensor, out0_tensor],
        [shape1_const, shape0_const]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_3d_input():
    """[3,4,4] float32 → reshape → [3,16] float32"""
    input_name = "input_3d"
    output_name = "output_3d"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [3, 4, 4]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [3, 16]
    )

    reshape_node = helper.make_node(
        "Reshape",
        inputs=[input_name, "shape_const"],
        outputs=[output_name],
    )

    shape_const = helper.make_tensor(
        name="shape_const",
        data_type=TensorProto.INT64,
        dims=[2],
        vals=[3, 16]
    )

    graph = helper.make_graph(
        [reshape_node],
        "3d_input",
        [input_tensor],
        [output_tensor],
        [shape_const]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_planar_chw():
    """[1,3,4,4] float32 → identity → [1,3,4,4] float32 (planar CHW layout)"""
    input_name = "input_chw"
    output_name = "output_chw"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [1, 3, 4, 4]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [1, 3, 4, 4]
    )

    identity = helper.make_node(
        "Identity",
        inputs=[input_name],
        outputs=[output_name],
    )

    graph = helper.make_graph(
        [identity],
        "planar_chw",
        [input_tensor],
        [output_tensor]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_invalid_1d():
    """[48] float32 → invalid 1D input"""
    input_name = "input_1d"
    output_name = "output_1d"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [48]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [48]
    )

    identity = helper.make_node("Identity", inputs=[input_name], outputs=[output_name])

    graph = helper.make_graph([identity], "invalid_1d", [input_tensor], [output_tensor])
    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_invalid_5d():
    """[1,2,2,2,3] float32 → invalid 5D input"""
    input_name = "input_5d"
    output_name = "output_5d"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [1, 2, 2, 2, 3]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [1, 2, 2, 2, 3]
    )

    identity = helper.make_node("Identity", inputs=[input_name], outputs=[output_name])

    graph = helper.make_graph([identity], "invalid_5d", [input_tensor], [output_tensor])
    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_invalid_channels_2():
    """[1,4,4,2] float32 → invalid 2-channel input"""
    input_name = "input_ch2"
    output_name = "output_ch2"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [1, 4, 4, 2]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [1, 4, 4, 2]
    )

    identity = helper.make_node("Identity", inputs=[input_name], outputs=[output_name])

    graph = helper.make_graph([identity], "invalid_ch2", [input_tensor], [output_tensor])
    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_multi_input():
    """Two [1,4,4,3] float32 inputs → Add → [1,4,4,3] output (invalid for onnxinference)"""
    input0_name = "input0"
    input1_name = "input1"
    output_name = "output_add"

    input0_tensor = helper.make_tensor_value_info(
        input0_name, TensorProto.FLOAT, [1, 4, 4, 3]
    )
    input1_tensor = helper.make_tensor_value_info(
        input1_name, TensorProto.FLOAT, [1, 4, 4, 3]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.FLOAT, [1, 4, 4, 3]
    )

    add_node = helper.make_node(
        "Add",
        inputs=[input0_name, input1_name],
        outputs=[output_name],
    )

    graph = helper.make_graph(
        [add_node],
        "multi_input",
        [input0_tensor, input1_tensor],
        [output_tensor]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_uint8_out():
    """[1,4,4,3] float32 → Cast UINT8 (unsupported output type)"""
    input_name = "input_f32"
    output_name = "output_u8"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [1, 4, 4, 3]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.UINT8, [1, 4, 4, 3]
    )

    cast_node = helper.make_node(
        "Cast",
        inputs=[input_name],
        outputs=[output_name],
        to=TensorProto.UINT8
    )

    graph = helper.make_graph(
        [cast_node],
        "uint8_out",
        [input_tensor],
        [output_tensor]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def make_model_float64_out():
    """[1,4,4,3] float32 → Cast FLOAT64 (unsupported output type)"""
    input_name = "input_f32"
    output_name = "output_f64"

    input_tensor = helper.make_tensor_value_info(
        input_name, TensorProto.FLOAT, [1, 4, 4, 3]
    )
    output_tensor = helper.make_tensor_value_info(
        output_name, TensorProto.DOUBLE, [1, 4, 4, 3]
    )

    cast_node = helper.make_node(
        "Cast",
        inputs=[input_name],
        outputs=[output_name],
        to=TensorProto.DOUBLE
    )

    graph = helper.make_graph(
        [cast_node],
        "float64_out",
        [input_tensor],
        [output_tensor]
    )

    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 11)])
    model.ir_version = 10
    return model


def main():
    # Clean old models
    for f in os.listdir(BASE_DIR):
        if f.endswith(('.onnx', '.modelinfo')):
            os.remove(os.path.join(BASE_DIR, f))

    # Valid models
    models = [
        ("flatten_float32in_float32out.onnx", make_model_flatten_f32_f32()),
        ("flatten_uint8in_float32out.onnx", make_model_uint8_to_f32()),
        ("int32out.onnx", make_model_int32_out()),
        ("dynamic_batch.onnx", make_model_dynamic_batch()),
        ("multi_output.onnx", make_model_multi_output()),
        ("flatten_3d_float32.onnx", make_model_3d_input()),
        ("planar_chw.onnx", make_model_planar_chw()),
        ("invalid_dims_1d.onnx", make_model_invalid_1d()),
        ("invalid_dims_5d.onnx", make_model_invalid_5d()),
        ("invalid_dims_channels_2.onnx", make_model_invalid_channels_2()),
        ("multi_input_two_tensors.onnx", make_model_multi_input()),
        ("uint8out.onnx", make_model_uint8_out()),
        ("float64out.onnx", make_model_float64_out()),
    ]

    for filename, model in models:
        path = os.path.join(BASE_DIR, filename)
        onnx.save(model, path)

        # For multi_output, mark output-1 (CHW tensor) as col-major
        if filename == "multi_output.onnx":
            write_modelinfo(path, per_output_dims_order={1: "col-major"})
        else:
            write_modelinfo(path)

    # Corrupt model
    with open(os.path.join(BASE_DIR, "corrupt_model.onnx"), "wb") as f:
        f.write(b"not-a-valid-onnx-model")

    print(f"Generated models in {BASE_DIR}")


if __name__ == "__main__":
    main()
